import sys
import textwrap
import unittest
from unittest import skipIf

from junkie.context import Context


class ContextTest(unittest.TestCase):
    def test_context_add_build(self):
        class Class:
            def __init__(self, prefix, suffix):
                self.text = prefix + suffix

        context = Context({
            "prefix": "abc",
            "suffix": "def",
            "text": lambda prefix, suffix: prefix + suffix
        })
        context.add(
            {"class": Class},
            {"suffix": "xyz"}
        )

        with context.build({"text", "prefix"}) as instance:
            self.assertEqual({"text": "abcxyz", "prefix": "abc"}, instance)

        with context.build(["text", "prefix"]) as instance:
            self.assertEqual({"text": "abcxyz", "prefix": "abc"}, instance)

        with context.build(("text", "prefix")) as (text, prefix):
            self.assertEqual("abcxyz", text)
            self.assertEqual("abc", prefix)

        with context.build("class") as instance:
            self.assertEqual("abcxyz", instance.text)

        with context.build(Class) as instance:
            self.assertEqual("abcxyz", instance.text)

    def test_add_none_context(self):
        context = Context({"value": "abc"}, None)
        context.add(None)

        with context.build("value") as value:
            self.assertEqual("abc", value)

    @skipIf(sys.version_info < (3, 7), "@dataclass needs at least Python 3.7")
    def test_dataclass_decorator(self):
        exec(textwrap.dedent("""
            from dataclasses import dataclass

            @dataclass
            class Class:
                prefix: str
                suffix: str
                pass
    
                def __post_init__(self):
                    self.text = self.prefix + self.suffix

            with Context({"prefix": "abc", "suffix": "def"}).build(Class) as instance:
                self.assertEqual("abcdef", instance.text)
        """))



from contextlib import contextmanager
from typing import Union, Dict, Callable, Set, List, Tuple

from junkie.core_context import CoreContext


class Context(CoreContext):
    def __init__(self, *instances_and_factories_args: Union[Dict[str, object], Dict[str, Callable], None]):
        super().__init__()

        self.add(*instances_and_factories_args)

    def add(self, *instances_and_factories_args: Union[Dict[str, object], Dict[str, Callable], None]):
        for instances_and_factories in instances_and_factories_args:
            if instances_and_factories is None:
                continue

            for key, value in instances_and_factories.items():
                if callable(value):
                    self._factories[key] = value
                else:
                    self._instances[key] = value

    @contextmanager
    def build(self, names_or_type: Union[Set[str], List[str], Tuple[str, ...], str, type]):
        if isinstance(names_or_type, (set, list)):
            with self.build_dict(names_or_type) as instance_dict:
                yield instance_dict

        elif isinstance(names_or_type, tuple):
            with self.build_instance_by_names(names_or_type) as instance:
                yield instance

        elif isinstance(names_or_type, str):
            with self.build_instance_by_name(names_or_type) as instance:
                yield instance

        elif isinstance(names_or_type, type):
            with self.build_instance_by_type(names_or_type) as instance:
                yield instance

        else:
            raise Exception("Type not known (set, list, str or type/class expected): {}".format(type(names_or_type)))



import inspect
import logging
from collections import OrderedDict
from contextlib import contextmanager, ExitStack
from typing import Union, Set, List, Dict, Callable, Tuple

import junkie


class CoreContext:
    def __init__(self):
        self.logger = logging.getLogger(junkie.__name__)

        self._instances = {}
        self._factories = {}

    def add_instances(self, instances: Dict[str, object]):
        self._instances.update(instances)

    def add_factories(self, factories: Dict[str, Callable]):
        self._factories.update(factories)

    @contextmanager
    def build_dict(self, names: Union[Set[str], List[str]]) -> Dict[str, object]:
        self.logger.debug("build(%s)", names)

        with ExitStack() as stack:
            instance_dict = {}

            for name in names:
                instance_dict[name] = self._build_instance_by_name(name, stack)

            yield instance_dict

    @contextmanager
    def build_instance_by_names(self, names: Tuple[str, ...]) -> Tuple[object]:
        self.logger.debug("build(%s)", names)

        with ExitStack() as stack:
            instances = []

            for name in names:
                instance = self._build_instance_by_name(name, stack)
                instances.append(instance)

            yield tuple(instances)

    @contextmanager
    def build_instance_by_name(self, name: str) -> object:
        self.logger.debug("build('%s')", name)

        with ExitStack() as stack:
            yield self._build_instance_by_name(name, stack)

    def _build_instance_by_name(self, name: str, stack: ExitStack):
        if name in self._instances:
            return self._instances[name]

        if name in self._factories:
            return self._call(self._factories[name], stack, name)

        raise Exception("Not found: {}".format(name))

    @contextmanager
    def build_instance_by_type(self, constructor: type) -> object:
        self.logger.debug("build(%s)", constructor.__name__)

        with ExitStack() as stack:
            yield self._call(constructor, stack, constructor.__name__)

    def _call(self, factory_func: Callable, stack: ExitStack, instance_name: str):
        args = OrderedDict()

        for name, annotation in inspect.signature(factory_func).parameters.items():
            if name in self._instances:
                arg = self._instances[name]

            elif name in self._factories:
                arg = self._call(self._factories[name], stack, name)

            elif annotation.default is not inspect.Parameter.empty:
                continue

            else:
                raise Exception("Not found: " + name)

            args[name] = arg

        self.logger.debug("%s = %s(%s)", instance_name, factory_func.__name__, list(args.keys()))
        instance = factory_func(**args)

        if hasattr(instance, "__enter__"):
            self.logger.debug("%s.__enter__()", instance_name)
            stack.push(lambda *exception_details: self.logger.debug("%s.__exit__()", instance_name))

            instance = stack.enter_context(instance)

        return instance



import logging
import random
import unittest
from contextlib import contextmanager

import junkie
from junkie.core_context import CoreContext


class CoreContextTest(unittest.TestCase):
    def test_simple_instance(self):
        context = CoreContext()
        context.add_instances({"text": "abc"})

        with context.build_dict({"text"}) as instances:
            self.assertEqual({"text": "abc"}, instances)

    def test_simple_factory(self):
        context = CoreContext()
        context.add_factories({"text": lambda: "abc"})

        with context.build_dict({"text"}) as instances:
            self.assertEqual({"text": "abc"}, instances)

    def test_factory_using_other_factory(self):
        context = CoreContext()
        context.add_instances({"prefix": "abc", "suffix": "def"})
        context.add_factories({"text": lambda prefix, suffix: prefix + suffix})

        with context.build_dict({"text", "prefix"}) as instances:
            self.assertEqual({"text": "abcdef", "prefix": "abc"}, instances)

    def test_build_instance_by_dict_with_list_in_right_order(self):
        def func(letter):
            def factory_func(logger):
                logger.append(letter)
                return letter

            return factory_func

        test_logger = []
        context = CoreContext()
        context.add_factories({"a": func("a"), "b": func("b"), "c": func("c"), "d": func("d"), "e": func("e")})
        context.add_instances({"logger": test_logger})

        names = ["a", "b", "c", "d", "e"]
        random.shuffle(names)

        with context.build_dict(names):
            self.assertEqual(names, test_logger)

    def test_build_instance_by_type(self):
        class Class:
            def __init__(self, prefix, suffix):
                self.text = prefix + suffix

        context = CoreContext()
        context.add_instances({"prefix": "abc", "suffix": "def"})

        with context.build_instance_by_type(Class) as instance:
            self.assertEqual("abcdef", instance.text)

    def test_build_instance_by_name(self):
        class Class:
            def __init__(self, prefix, suffix):
                self.text = prefix + suffix

        context = CoreContext()
        context.add_instances({"prefix": "abc", "suffix": "def"})
        context.add_factories({"text": Class})

        with context.build_instance_by_name("text") as instance:
            self.assertEqual("abcdef", instance.text)

    def test_build_instance_by_names(self):
        class Class:
            def __init__(self, prefix, suffix):
                self.text = prefix + suffix

        context = CoreContext()
        context.add_instances({"prefix": "abc", "suffix": "def"})
        context.add_factories({"text": Class})

        with context.build_instance_by_names(("text", "text")) as (instance1, instance2):
            self.assertEqual("abcdef", instance1.text)
            self.assertEqual("abcdef", instance2.text)

            self.assertNotEqual(instance1, instance2)

    def test_default_argument_usage(self):
        class MyClassWithDefaultArguments:
            def __init__(self, argument: str, default_argument: int = 10, default_argument2: str = None):
                self.argument = argument
                self.default_argument = default_argument
                self.default_argument2 = default_argument2 or "Hello"

        core_context = CoreContext()
        core_context.add_instances({"argument": "value"})

        with core_context.build_instance_by_type(MyClassWithDefaultArguments) as instance:
            self.assertEqual("value", instance.argument)
            self.assertEqual(10, instance.default_argument)
            self.assertEqual("Hello", instance.default_argument2)

    def test_partial_default_arguments_usage(self):
        class MyClassWithDefaultArguments:
            def __init__(self, argument: str, default_argument: int = 10, default_argument2: str = None):
                self.argument = argument
                self.default_argument = default_argument
                self.default_argument2 = default_argument2 or "Hello"

        core_context = CoreContext()
        core_context.add_instances({"argument": "value", "default_argument2": "set from context"})

        with core_context.build_instance_by_type(MyClassWithDefaultArguments) as instance:
            self.assertEqual("value", instance.argument)
            self.assertEqual(10, instance.default_argument)
            self.assertEqual("set from context", instance.default_argument2)

    def test_context_manager_enter_and_exit(self):
        class Class:
            def __init__(self, message_service, database_context):
                self.message_service = message_service
                self.database_context = database_context

        class MessageService:
            def __init__(self, logger):
                self.logger = logger

            def __enter__(self):
                self.logger.append("connect")
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.logger.append("disconnect")

        @contextmanager
        def create_database(logger):
            logger.append("open")
            yield "DB"
            logger.append("close")

        test_logger = list()
        context = CoreContext()
        context.add_instances({"logger": test_logger})
        context.add_factories({"message_service": MessageService, "database_context": create_database})

        with context.build_instance_by_type(Class) as instance:
            self.assertEqual(MessageService, type(instance.message_service))
            self.assertEqual("DB", instance.database_context)

            self.assertEqual(["connect", "open"], test_logger)

        self.assertEqual(["connect", "open", "close", "disconnect"], test_logger)

    def test_logging(self):
        self.maxDiff = None

        class Class:
            def __init__(self, database_context, message_service, text):
                self.database_context = database_context
                self.message_service = message_service
                self.text = text

        @contextmanager
        def create_database(connection_string):
            yield "DB: " + connection_string

        @contextmanager
        def create_message_service():
            yield "message-service"

        context = CoreContext()
        context.add_instances({"text": "abc"})
        context.add_factories({
            "connection_string": lambda: "URL",
            "database_context": create_database,
            "message_service": create_message_service,
            "class": Class,
        })

        with self.assertLogs(level="DEBUG") as logging_context:
            with context.build_instance_by_type(Class):
                logging.getLogger(__name__).info("execute context block")

        self.assertEqual([
            "DEBUG:{}:build(Class)".format(junkie.__name__),
            "DEBUG:{}:connection_string = <lambda>([])".format(junkie.__name__),
            "DEBUG:{}:database_context = create_database(['connection_string'])".format(junkie.__name__),
            "DEBUG:{}:database_context.__enter__()".format(junkie.__name__),
            "DEBUG:{}:message_service = create_message_service([])".format(junkie.__name__),
            "DEBUG:{}:message_service.__enter__()".format(junkie.__name__),
            "DEBUG:{}:Class = Class(['database_context', 'message_service', 'text'])".format(junkie.__name__),
            "INFO:{}:execute context block".format(__name__),
            "DEBUG:{}:message_service.__exit__()".format(junkie.__name__),
            "DEBUG:{}:database_context.__exit__()".format(junkie.__name__),
        ], logging_context.output)
